#ifndef _NMS_ROTATED_
#define _NMS_ROTATED_

// <<<============================= Header files =============================>>>
#include <algorithm>
#include <vector>
#include <array>
#include <cmath>
#include <cassert>
#include <numeric>
#include "Eigen/Dense"

// <<<============================= Constants =============================>>>
static constexpr size_t BOX_SIZE{5}; // either x1, y1, x2, y2, r or x1, y1, w, h, r

// <<<============================= Type definitions =============================>>>
template<typename T>
struct RotatedBox_t{
    T x_ctr, y_ctr, w, h, a;
};

template<typename T>
struct Point_t{
    T x, y;

    inline Point_t(const T& px = 0, const T& py = 0) : x{px}, y{py} {}
    inline Point_t operator+(const Point_t& p) const { return Point_t{x+p.x, y+p.y}; }
    inline Point_t& operator+=(const Point_t& p){
        x += p.x;
        y += p.y;
        return *this;
    }
    inline Point_t operator-(const Point_t& p) const { return Point_t{x-p.x, y-p.y}; }
    inline Point_t operator*(const T coeff) const { return Point_t{x*coeff, y*coeff}; }
};

// <<<============================= Type aliases =============================>>>
template<typename T>
using RotatedBox = RotatedBox_t<T>;

template<typename T>
using Point = Point_t<T>;

// A type showing a list of boxes used in
// NMS functions - The context shows whether
// the boxes are represented as xyxyr or xywhr
template<typename T>
using Boxes = std::vector<std::array<T, BOX_SIZE>>;

// <<<============================= Functions =============================>>>
// Based on https://stackoverflow.com/questions/1577475/c-sorting-and-keeping-track-of-indexes
template<typename T>
inline std::vector<int> sort_indices(const Eigen::VectorX<T>& v, const bool desc = true){
    std::vector<int> idx(v.size());
    std::iota(idx.begin(), idx.end(), 0);

    if(desc)
        std::stable_sort(idx.begin(), idx.end(),
                [&v](int i1, int i2) {return v[i1] > v[i2];}
                );
    else
        std::stable_sort(idx.begin(), idx.end(),
                [&v](int i1, int i2) {return v[i1] < v[i2];}
                );
    return idx;
}

template<typename T>
inline void get_rotated_vertices(
    const RotatedBox<T>& box,
    Point<T> (&pts)[4])
{
    // M_PI / 180. == 0.01745329251
    // double theta = box.a * 0.01745329251;
    // MODIFIED    double theta = box.a;
    double theta = box.a;
    T cos_theta2 = static_cast<T>(cos(theta) * 0.5f);
    T sin_theta2 = static_cast<T>(sin(theta) * 0.5f);

    // y: top --> down; x: left --> right
    pts[0].x = box.x_ctr - sin_theta2 * box.h - cos_theta2 * box.w;
    pts[0].y = box.y_ctr + cos_theta2 * box.h - sin_theta2 * box.w;
    pts[1].x = box.x_ctr + sin_theta2 * box.h - cos_theta2 * box.w;
    pts[1].y = box.y_ctr - cos_theta2 * box.h - sin_theta2 * box.w;
    pts[2].x = 2 * box.x_ctr - pts[0].x;
    pts[2].y = 2 * box.y_ctr - pts[0].y;
    pts[3].x = 2 * box.x_ctr - pts[1].x;
    pts[3].y = 2 * box.y_ctr - pts[1].y;
}

template<typename T>
inline T cross_2d(const Point<T>& A, const Point<T>& B){
    return A.x*B.y - B.x*A.y;
}

template<typename T>
inline T dot_2d(const Point<T>& A, const Point<T>& B){
    return A.x*B.x + A.y*B.y;
}

template<typename T>
int get_intersection_points(
    const Point<T> (&pts1)[4],
    const Point<T> (&pts2)[4],
    Point<T> (&intersections)[24])
{
    // Line vector
    // A line from p1 to p2 is: p1 + (p2-p1)*t, t=[0,1]
    Point<T> vec1[4], vec2[4];
    for (int i = 0; i < 4; i++) {
        vec1[i] = pts1[(i + 1) % 4] - pts1[i];
        vec2[i] = pts2[(i + 1) % 4] - pts2[i];
    }

    // Line test - test all line combos for intersection
    int num = 0;  // number of intersections
    for (int i = 0; i < 4; i++) {
        for (int j = 0; j < 4; j++) {
            // Solve for 2x2 Ax=b
            T det = cross_2d<T>(vec2[j], vec1[i]);

            // This takes care of parallel lines
            if (fabs(det) <= 1e-14) {
                continue;
            }

            auto vec12 = pts2[j] - pts1[i];

            T t1 = cross_2d<T>(vec2[j], vec12) / det;
            T t2 = cross_2d<T>(vec1[i], vec12) / det;

            if (t1 >= 0.0f && t1 <= 1.0f && t2 >= 0.0f && t2 <= 1.0f) {
                intersections[num++] = pts1[i] + vec1[i] * t1;
            }
        }
    }

    // Check for vertices of rect1 inside rect2
    {
        const auto& AB = vec2[0];
        const auto& DA = vec2[3];
        auto ABdotAB = dot_2d<T>(AB, AB);
        auto ADdotAD = dot_2d<T>(DA, DA);
        for (int i = 0; i < 4; i++) {
            // assume ABCD is the rectangle, and P is the point to be judged
            // P is inside ABCD iff. P's projection on AB lies within AB
            // and P's projection on AD lies within AD

            auto AP = pts1[i] - pts2[0];

            auto APdotAB = dot_2d<T>(AP, AB);
            auto APdotAD = -dot_2d<T>(AP, DA);

            if ((APdotAB >= 0) && (APdotAD >= 0) && (APdotAB <= ABdotAB) &&
                (APdotAD <= ADdotAD)) {
                intersections[num++] = pts1[i];
            }
        }
    }

    // Reverse the check - check for vertices of rect2 inside rect1
    {
        const auto& AB = vec1[0];
        const auto& DA = vec1[3];
        auto ABdotAB = dot_2d<T>(AB, AB);
        auto ADdotAD = dot_2d<T>(DA, DA);
        for (int i = 0; i < 4; i++) {
            auto AP = pts2[i] - pts1[0];
            auto APdotAB = dot_2d<T>(AP, AB);
            auto APdotAD = -dot_2d<T>(AP, DA);
            if ((APdotAB >= 0) && (APdotAD >= 0) && (APdotAB <= ABdotAB) &&
                (APdotAD <= ADdotAD)) {
                intersections[num++] = pts2[i];
            }
        }
    }

  return num;
}

template<typename T>
int convex_hull_graham(
    const Point<T> (&p)[24],
    const int& num_in, Point<T> (&q)[24],
    bool shift_to_zero = false)
{
    assert(num_in >= 2);

    // Step 1:
    // Find point with minimum y
    // if more than 1 points have the same minimum y,
    // pick the one with the minimum x.
    int t = 0;
    for (int i = 1; i < num_in; i++) {
        if (p[i].y < p[t].y || (p[i].y == p[t].y && p[i].x < p[t].x)) {
            t = i;
        }
    }
    auto& start = p[t];  // starting point
    // Step 2:
    // Subtract starting point from every points (for sorting in the next step)
    for (int i = 0; i < num_in; i++) {
        q[i] = p[i] - start;
    }
    // Swap the starting point to position 0
    auto tmp = q[0];
    q[0] = q[t];
    q[t] = tmp;
    // Step 3:
    // Sort point 1 ~ num_in according to their relative cross-product values
    // (essentially sorting according to angles)
    // If the angles are the same, sort according to their distance to origin
    T dist[24];
    for (int i = 0; i < num_in; i++) {
        dist[i] = dot_2d<T>(q[i], q[i]);
    }

    // CPU version
    std::sort(q + 1, q + num_in,
              [](const Point<T>& A, const Point<T>& B) -> bool {
                T temp = cross_2d<T>(A, B);
                if (fabs(temp) < 1e-6) {
                    return dot_2d<T>(A, A) < dot_2d<T>(B, B);
                } else {
                    return temp > 0;
                }
            });
    // compute distance to origin after sort, since the points are now different.
    for (int i = 0; i < num_in; i++) {
        dist[i] = dot_2d<T>(q[i], q[i]);
    }

    // Step 4:
    // Make sure there are at least 2 points (that don't overlap with each other)
    // in the stack
    int k;  // index of the non-overlapped second point
    for (k = 1; k < num_in; k++) {
        if (dist[k] > 1e-8) {
            break;
        }
    }
    if (k == num_in) {
        // We reach the end, which means the convex hull is just one point
        q[0] = p[t];
        return 1;
    }
    q[1] = q[k];
    int m = 2;  // 2 points in the stack
    // Step 5:
    // Finally we can start the scanning process.
    // When a non-convex relationship between the 3 points is found
    // (either concave shape or duplicated points),
    // we pop the previous point from the stack
    // until the 3-point relationship is convex again, or
    // until the stack only contains two points
    for (int i = k + 1; i < num_in; i++) {
        while (m > 1 && cross_2d<T>(q[i] - q[m - 2], q[m - 1] - q[m - 2]) >= 0) {
            m--;
        }
        q[m++] = q[i];
    }

    // Step 6 (Optional):
    // In general sense we need the original coordinates, so we
    // need to shift the points back (reverting Step 2)
    // But if we're only interested in getting the area/perimeter of the shape
    // We can simply return.
    if (!shift_to_zero) {
        for (int i = 0; i < m; i++) {
            q[i] += start;
        }
    }
    return m;
}

template<typename T>
T polygon_area(const Point<T> (&q)[24], const int& m)
{
    if (m <= 2) {
        return 0;
    }

    T area = 0;
    for (int i = 1; i < m - 1; i++) {
        area += fabs(cross_2d<T>(q[i] - q[0], q[i + 1] - q[0]));
    }

    return area / 2.0;
}

template<typename T>
T rotated_boxes_intersection(
    const RotatedBox<T>& box1,
    const RotatedBox<T>& box2)
{
    // There are up to 4 x 4 + 4 + 4 = 24 intersections (including dups) returned
    // from rotated_rect_intersection_pts
    Point<T> intersectPts[24], orderedPts[24];

    Point<T> pts1[4];
    Point<T> pts2[4];
    get_rotated_vertices<T>(box1, pts1);
    get_rotated_vertices<T>(box2, pts2);

    int num = get_intersection_points<T>(pts1, pts2, intersectPts);
    if (num <= 2) {
        return 0.0;
    }

    // Convex Hull to order the intersection points in clockwise order and find
    // the contour area.
    int num_convex = convex_hull_graham<T>(intersectPts, num, orderedPts, true);
    return polygon_area<T>(orderedPts, num_convex);
}

template<typename T>
T single_box_iou_rotated(
    const std::vector<T>& box1_raw,
    const std::vector<T>& box2_raw,
    const int& mode_flag)
{
    // shift center to the middle point to achieve higher precision in result
    RotatedBox<T> box1, box2;
    auto center_shift_x = (box1_raw[0] + box2_raw[0]) / 2.0;
    auto center_shift_y = (box1_raw[1] + box2_raw[1]) / 2.0;
    box1.x_ctr = box1_raw[0] - center_shift_x;
    box1.y_ctr = box1_raw[1] - center_shift_y;
    box1.w = box1_raw[2];
    box1.h = box1_raw[3];
    box1.a = box1_raw[4];
    box2.x_ctr = box2_raw[0] - center_shift_x;
    box2.y_ctr = box2_raw[1] - center_shift_y;
    box2.w = box2_raw[2];
    box2.h = box2_raw[3];
    box2.a = box2_raw[4];

    const T area1 = box1.w * box1.h;
    const T area2 = box2.w * box2.h;
    if (area1 < 1e-14 || area2 < 1e-14) {
        return 0.f;
    }

    const T intersection = rotated_boxes_intersection<T>(box1, box2);
    T baseS = 1.0;
    if (mode_flag == 0) {
        baseS = (area1 + area2 - intersection);
    } else if (mode_flag == 1) {
        baseS = area1;
    }

    const T iou = intersection / baseS;
    return iou;
}

// Main function
//dets: Each element is of size 5 and shows a detection
template<typename T>
std::vector<int> nms_rotated(
    const Matrix_R<T>& dets,
    const Eigen::VectorX<T>& scores,
    const float& iou_threshold)
{
    if(dets.size() == 0)
        return std::vector<int>();

    auto order_t = sort_indices(scores, true);
    auto ndets = dets.rows();
    std::vector<bool> suppressed_t(ndets, false);
    std::vector<int> keep_t(ndets, 0);

    int num_to_keep{0};
    for(int _i = 0; _i < ndets; _i++){
        auto i = order_t[_i];
        if(suppressed_t[i] == 1)
            continue;

        keep_t[num_to_keep++] = i;

        for(int _j = _i + 1; _j < ndets; _j++){
            auto j = order_t[_j];
            if(suppressed_t[j] == 1)
                continue;

            std::vector<T> det_i(dets.row(i).data(), dets.row(i).data()+dets.cols());
            std::vector<T> det_j(dets.row(j).data(), dets.row(j).data()+dets.cols());
            auto ovr = single_box_iou_rotated<T>(
                det_i, det_j, 0);
            if(ovr >= iou_threshold)
                suppressed_t[j] = 1;
        }
    }
    return std::vector<int>(keep_t.begin(), keep_t.begin()+num_to_keep);
}

template<typename T>
std::vector<int> nms_bev(
    const Matrix_R<T>& boxes,
    const Eigen::VectorX<T>& scores,
    const float thresh)
{
    auto order = sort_indices(scores, true);
    // Change order of boxes and scores
    Matrix_R<T> boxes_reordered(boxes.rows(), boxes.cols());
    Eigen::VectorX<T> scores_reordered(scores.size());
    for(int i = 0; i < scores.rows(); i++){
        boxes_reordered.row(i) = boxes.row(order[i]);
        scores_reordered[i] = scores[order[i]];
    }
    Matrix_R<T> boxes_stacked(boxes_reordered.rows(), boxes_reordered.cols());
    boxes_stacked.col(0) = (boxes_reordered.col(0)+boxes_reordered.col(2))/2;
    boxes_stacked.col(1) = (boxes_reordered.col(1)+boxes_reordered.col(3))/2;
    boxes_stacked.col(2) = boxes_reordered.col(2)-boxes_reordered.col(0);
    boxes_stacked.col(3) = boxes_reordered.col(3)-boxes_reordered.col(1);
    boxes_stacked.col(4) = boxes_reordered.col(4);

    auto keep = nms_rotated(boxes_stacked, scores_reordered, thresh);
//    auto keep = nms_rotated(boxes_reordered, scores_reordered, thresh); //nima code

    std::vector<int> keep_reordered(keep.size());
    for(int i = 0; i < keep.size(); i++)
        keep_reordered[i] = order[keep[i]];

    return keep_reordered;
}

template<typename T>
MulticlassRes<T> box3d_multiclass_nms(
    const Matrix_R<T>& mlvl_bboxes,
    const Matrix_R<T>& mlvl_bboxes_for_nms,
    const Matrix_R<T>& mlvl_scores,
    const float& score_thr,
    const int& max_num,
    const Eigen::VectorXi& mlvl_dir_scores)
{
    auto num_classes = mlvl_scores.cols() - 1;

    Vec<Matrix_R<T>> bboxes;
    Vec<Eigen::VectorX<T>> scores;
    Vec<Eigen::VectorXi> labels;
    Vec<Eigen::VectorXi> dir_scores;
    MulticlassRes<T> results;

    int bboxes_cols{0}, bboxes_rows{0};
    for(int i = 0; i < num_classes; i++){
        Eigen::VectorXi cls_inds = (mlvl_scores.col(i).array() > static_cast<T>(score_thr)).template cast<int>();
        Eigen::VectorX<T> _scores = keep_if(static_cast<Eigen::VectorX<T>>(mlvl_scores.col(i)), cls_inds);
        Matrix_R<T> _bboxes_for_nms = keep_if(mlvl_bboxes_for_nms, cls_inds);
        Matrix_R<T> _mlvl_bboxes = keep_if(mlvl_bboxes, cls_inds);
        Eigen::VectorXi _mlvl_dir_scores = keep_if(mlvl_dir_scores, cls_inds);
        std::vector<int> selected = nms_bev(_bboxes_for_nms, _scores, 0.8f);

        bboxes_rows += selected.size();
        bboxes.push_back(_mlvl_bboxes(selected, Eigen::all));
        scores.push_back(_scores(selected, Eigen::all));
        labels.push_back(Eigen::VectorXi(selected.size()).setConstant(i));
        dir_scores.push_back(_mlvl_dir_scores(selected, Eigen::all));
    }
    if (bboxes.size()){
        std::cout << "Combining results...\n";
        Matrix_R<T> bboxes_all = cat(bboxes);
        Eigen::VectorX<T> scores_all = cat(scores);
        Eigen::VectorXi labels_all = cat(labels);
        Eigen::VectorXi dir_scores_all = cat(dir_scores);

        if (bboxes_all.rows() > max_num){
            auto inds = sort_indices(scores_all, true);
            std::vector<int> first200(inds.begin(), inds.begin() + max_num);
            results.bboxes = bboxes_all(first200, Eigen::all);
            results.scores = scores_all(first200);
            results.labels = labels_all(first200);

            //nima code
//            results.bboxes = bboxes_all(inds, Eigen::all);
//            results.scores = scores_all(inds);
//            results.labels = labels_all(inds);

            results.dir_scores = dir_scores_all(first200);
        } else {
            results.bboxes = bboxes_all;
            results.scores = scores_all;
            results.labels = labels_all;
            results.dir_scores = dir_scores_all;
        }
    }
    return results;
}

#endif //_NMS_ROTATED_
